package gg.malloc.defense;

import org.bukkit.plugin.java.JavaPlugin;
import org.bukkit.plugin.RegisteredServiceProvider;
import org.bukkit.Location;
import org.bukkit.entity.Entity;
import org.bukkit.entity.EntityType;
import org.bukkit.entity.Player;
import org.bukkit.World;
import org.bukkit.WorldCreator;

import org.bukkit.event.Listener;
import org.bukkit.event.player.PlayerQuitEvent;
import org.bukkit.event.EventHandler;

import java.util.logging.Logger;
import java.util.logging.Level;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Collection;

import gg.malloc.defense.model.Arena;
import gg.malloc.defense.model.Game;

import gg.malloc.defense.games.LinearGame;
import gg.malloc.defense.games.ScaledWaves;

import gg.malloc.defense.engine.GameRunner;

import gg.malloc.defense.config.Configuration;

import gg.malloc.defense.commands.AddPlayerCommand;
import gg.malloc.defense.commands.SetStageCommand;
import gg.malloc.defense.commands.PlayerReadyCommand;
import gg.malloc.defense.commands.ListGamesCommand;
import gg.malloc.defense.commands.JoinGameCommand;
import gg.malloc.defense.commands.LeaveGameCommand;
import gg.malloc.defense.commands.ReloadCommand;

public class Plugin extends JavaPlugin {
  HashMap<String, Arena> m_arenas = new HashMap<>();
  ArrayList<Game> m_games = new ArrayList<>();
  HashMap<String, GameRunner> m_runningGames = new HashMap<>();
  HashMap<Player, GameRunner> m_playerGames = new HashMap<>();
  GameEventHandler m_handler = new GameEventHandler();

  public Collection<String> arenaNames() {
    return m_arenas.keySet();
  }

  public void debuginfo() {
    getLogger().info("Debug Info:");
  }

  @Override
  public void onEnable() {
    getLogger().info("Malloc Defense registered");
    getLogger().setLevel(Level.FINEST);
    m_games.add(new ScaledWaves());
    reloadArenas();
    getCommand("join").setExecutor(new JoinGameCommand(this));
    getCommand("leave").setExecutor(new LeaveGameCommand(this));
    getCommand("ready").setExecutor(new PlayerReadyCommand(this));
    getCommand("list").setExecutor(new ListGamesCommand(this));

    getCommand("addplayer").setExecutor(new AddPlayerCommand(this));
    getCommand("setstage").setExecutor(new SetStageCommand(this));
    getCommand("debuginfo").setExecutor(new DebuginfoCommand(this));
    getCommand("reload").setExecutor(new ReloadCommand(this));

    getServer().getPluginManager().registerEvents(new PlayerQuitHandler(), this);
    getServer().getPluginManager().registerEvents(m_handler, this);
  }

  public void reloadArenas() {
    getLogger().info("Loading arenas...");
    saveDefaultConfig();
    reloadConfig();
    Configuration config = new Configuration(getConfig());

    for(GameRunner runner : m_runningGames.values()) {
      runner.requestTransition(GameRunner.Stage.Idle);
    }
    for(Player player : m_playerGames.keySet()) {
      m_playerGames.get(player).removePlayer(player);
    }
    m_runningGames.clear();
    m_playerGames.clear();
    m_handler.clear();
    m_arenas.clear();
    for(String mapName : config.getMapNames()) {
      m_arenas.put(mapName, config.getArenaForMap(mapName));
    }
  }

  public boolean hasRunnerForArenaName(String arenaName) {
    return m_runningGames.containsKey(arenaName);
  }

  public void addPlayerToArena(String arenaName, Player player) {
    GameRunner runner = getRunnerForArenaName(arenaName);
    runner.addPlayer(player);
    m_playerGames.put(player, runner);
  }

  class PlayerQuitHandler implements Listener {
    @EventHandler
    public void onPlayerQuit(PlayerQuitEvent evt) {
      m_playerGames.remove(evt.getPlayer());
    }
  }

  public GameRunner getRunnerForArenaName(String arenaName) {
    GameRunner ret = null;
    if (m_runningGames.containsKey(arenaName)) {
      ret = m_runningGames.get(arenaName);
    } else if (m_arenas.containsKey(arenaName)) {
      getLogger().info("Loading game world " + arenaName);
      World gameWorld = getServer().getWorld(arenaName);
      if (gameWorld == null) {
        getLogger().info("Creating game world " + arenaName);
        gameWorld = new WorldCreator(arenaName).generateStructures(false).createWorld();
      }
      ret = new GameRunner(this, m_games.get(0), m_arenas.get(arenaName), gameWorld);
      m_handler.addRunner(ret);
      m_runningGames.put(arenaName, ret);
      getLogger().info("Game ready for " + arenaName);
    }
    return ret;
  }

  public GameRunner getRunnerForPlayer(Player p) {
    GameRunner ret = null;
    if (m_playerGames.containsKey(p)) {
      ret = m_playerGames.get(p);
    }
    return ret;
  }
}
